using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Text;
using System.Windows.Forms;

using TangentAPI;

namespace ChatApplication
{
    /// <summary>
    /// ChatApplication Notes:
    /// 
    /// To run, simply run the executable, then click the Server radio button.  Then
    /// from another machine (or even from the same machine), run another instance of
    /// the executable and click the Client radio button.  Messages can then be sent
    /// back and forth between these two running applications.  More applications
    /// can be started as Clients.  Using the Demo version of Tangent API, up to three
    /// clients can connect to the server simultaneously.
    /// 
    /// Message can only be sent from the server to all clients, and from the client
    /// to the server.  With a little bit more development effort, messages could be
    /// sent from one client to another client, by routing the message through the
    /// server.  See the Tangent API Model on www.tangentapi.com to understand better
    /// how Tangent API can be used to develop a solution for your needs.
    /// 
    /// This project uses Tangent API as an isolated COM reference.  This means, when
    /// this project is deployed to a computer, Tangent API does not need to be
    /// registered on that machine.  However, in the folder where ChatApplication.exe
    /// is, the files ChatApplication.exe.manifest, Interop.TangentAPI.dll, and
    /// TangentAPI.dll must also be present.  Tangent API was added as an isolated
    /// reference by first adding the Tangent API COM reference then setting the
    /// Isolated option of this reference to True.  Note that Tangent API must be
    /// registered on the development machine (in order to have added Tangent API as an
    /// isolated reference and to then build the project).  However, again, on the
    /// deployment machine, Tangent API does not need to be registered.
    /// 
    /// Read the comments throughout the code to better understand how this application
    /// works.
    /// </summary>
    public partial class Form1 : Form
    {
        #region Server methods and data
        TAServer _server = null;
        string _ipAddress;

        void RunAsServer()
        {
            // Create a new server object
            _server = new TAServer();

            // Register for server events
            _server.ReceivedDataFromClient += new _ITAServerEvents_ReceivedDataFromClientEventHandler(_server_ReceivedDataFromClient);
            _server.ConnectedClient += new _ITAServerEvents_ConnectedClientEventHandler(_server_ConnectedClient);
            _server.DisconnectedClient += new _ITAServerEvents_DisconnectedClientEventHandler(_server_DisconnectedClient);

            // Get all IP addresses on the system
            string[] ipAddresses = ObjectArrayToStringArray((object[])_server.GetIPAddresses());

            // Use the first IP address in the list.
            // Todo: If you want to use a different IP address for the chat server, edit
            // this line of code.
            _ipAddress = ipAddresses[0];

            // Startup the server
            _server.Startup(_ipAddress, PORT, 10);

            // Broadcast this server's IP address so that clients can
            // find this IP address automatically.  This makes it so that client don't
            // have to have a preconfigured IP address.
            _server.StartBroadcastingIPAddress(PORT);
        }

        /// <summary>
        /// This is the event handler for when data is received from
        /// a client.
        /// </summary>
        void _server_ReceivedDataFromClient(int ClientID, object data)
        {
            HandleReceivedData(data);
        }

        /// <summary>
        /// This is the event handler for when a client has connected.
        /// </summary>
        void _server_ConnectedClient(int ClientID)
        {
            AppendMessage("Client " + ClientID + " connected.");
        }

        /// <summary>
        /// This is the event handler for when a client has disconnected.
        /// </summary>
        void _server_DisconnectedClient(int ClientID)
        {
            AppendMessage("Client " + ClientID + " disconnected.");
        }
        #endregion

        #region Client methods and data
        TAClient _client = null;

        void RunAsClient()
        {
            // Create a new client object
            _client = new TAClient();

            // Register for client events
            _client.DisconnectedFromServer += new _ITAClientEvents_DisconnectedFromServerEventHandler(_client_DisconnectedFromServer);
            _client.ReceivedDataFromServer += new _ITAClientEvents_ReceivedDataFromServerEventHandler(_client_ReceivedDataFromServer);

            // Get the server's IP address.  The server is assumed to be broadcasting
            // its IP address.  Note, the IP address does not need to be configured
            // by the client.  The client simply obtains the server's IP address, then
            // uses the IP address to connect.  This is what it means to create a 
            // "zero-configuration client" regarding Tangent API.
            _ipAddress = _client.GetServerIPAddress(PORT);

            // Connect to the server, using the obtained IP address.
            _client.ConnectToServer(_ipAddress, PORT);
        }

        /// <summary>
        /// This is the event handler for when this client has received
        /// data from the server.
        /// </summary>
        void _client_ReceivedDataFromServer(object data)
        {
            HandleReceivedData(data);
        }

        /// <summary>
        /// This is the event handler for when the server has disconnected.
        /// </summary>
        void _client_DisconnectedFromServer()
        {
            AppendMessage("Server disconnected.");
        }
        #endregion

        #region Common methods and data
        const short PORT = 2550;
        /// <summary>
        /// This is the data structure that will be sent/received
        /// between the server and clients.
        /// </summary>
        [Serializable]
        struct DataToSend
        {
            public string Name;
            public string Text;
        }
        /// <summary>
        /// Display a message box for when exceptions are thrown.
        /// </summary>
        void HandleError(Exception ex)
        {
            MessageBox.Show(this, ex.Message, "Error occurred", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
        /// <summary>
        /// This method converts an object array to a string array.  This helper method comes in
        /// handy for the returned object from _server.GetIPAddresses().
        /// </summary>
        string[] ObjectArrayToStringArray(object[] objectArray)
        {
            string[] stringArray = new string[objectArray.Length];
            objectArray.CopyTo(stringArray, 0);
            return stringArray;
        }
        /// <summary>
        /// This method converts an object array to a int array.  This helper method comes in
        /// handy for the returned object from _server.GetConnectedClientIDs().
        /// </summary>
        int[] ObjectArrayToIntArray(object[] objectArray)
        {
            int[] intArray = new int[objectArray.Length];
            objectArray.CopyTo(intArray, 0);
            return intArray;
        }
        /// <summary>
        /// This is a rudimentary way to check if we're running as a server or a client
        /// </summary>
        bool RunningAsServer()
        {
            if (_server != null)
                return true;
            else
                return false;
        }
        /// <summary>
        /// This method handles data, and updates the GUI accordingly.  This method
        /// has been implemented to work the same whether a client has received
        /// data or whether a server has received data.  But in a more sophisticated
        /// project, the server and client may want to handle data differently.
        /// </summary>
        void HandleReceivedData(object data)
        {
            try
            {
                if (data is byte[])
                {
                    // First convert the bytes to an object
                    object o = ConvertBytesToObject((byte[])data);

                    if (o is DataToSend)
                    {
                        // Received object of type DataToSend
                        DataToSend dataToSend = (DataToSend)o;

                        AppendMessage("<" + dataToSend.Name + "> " + dataToSend.Text);
                    }
                    else
                    {
                        // Todo, handle any other object data types
                    }
                }
                else
                {
                    // Todo, handle any other data types received from clients
                }
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
        }
        /// <summary>
        /// This method converts a serializable object to bytes.
        /// </summary>
        byte[] ConvertObjectToBytes(object objectToConvert)
        {

            MemoryStream memoryStream = new MemoryStream();
            BinaryFormatter binaryFormatter = new BinaryFormatter();

            // If this method throws an exception, make sure the object you passed in is serializable
            binaryFormatter.Serialize(memoryStream, objectToConvert);

            byte[] objectBytes = memoryStream.ToArray();

            return objectBytes;
        }
        /// <summary>
        /// This method converts bytes to an object.
        /// </summary>
        object ConvertBytesToObject(byte[] bytesToConvert)
        {
            MemoryStream memoryStream = new MemoryStream(bytesToConvert);
            BinaryFormatter binaryFormatter = new BinaryFormatter();
            object convertedObject = binaryFormatter.Deserialize(memoryStream);
            return convertedObject;
        }
        /// <summary>
        /// This method sends a text message
        /// </summary>
        void SendMessage(string message)
        {
            // Create a data object to send
            DataToSend dataToSend;
            dataToSend.Name = RunningAsServer() ? "Server" : "Client " + _client.ClientID;
            dataToSend.Text = message;

            // Convert the data object to bytes
            byte[] objectDataBytes = ConvertObjectToBytes(dataToSend);

            // Encrypt the data
            byte[] dataBytesToSend = EncryptData(objectDataBytes);

            // Now send the data
            if (RunningAsServer())
            {
                // Running as server, so send to all clients
                _server.SendDataToAllClients(dataBytesToSend);

                // Todo: With a little more work, data could be sent to specific
                // clients by using the SendDataToClient method.  You can use the
                // clientID to specify which client to send data to when using
                // this method.
            }
            else
            {
                // Running as client, so send to the server
                _client.SendDataToServer(dataBytesToSend);
            }

            // NOTE: Above, we're sending byte[] via the Tangent API SendData...
            // methods.  Remember that you can also send many other object types
            // including string, int, bool, object[], etc.  Refer to the Tangent
            // API documentation to see the various object types that can be sent.
        }
        /// <summary>
        /// Encrypt the data, if desired
        /// </summary>
        /// <param name="dataBytes"></param>
        /// <returns></returns>
        byte[] EncryptData(byte[] dataBytes)
        {
            // Todo: If encryption of these bytes is desired, it
            // may be done here.

            return dataBytes;
        }
        #endregion

        #region GUI methods
        /// <summary>
        /// Appends a message to the textbox
        /// </summary>
        void AppendMessage(string message)
        {
            textBox2.Text += message + "\r\n";
            textBox1.Focus();
        }

        public Form1()
        {
            InitializeComponent();
        }

        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {
            Cursor.Current = Cursors.WaitCursor;

            // Attempt to run chat server
            try
            {
                EnableChatControls(true);

                RunAsServer();

                this.Text = "Chat Application - Server running on " + _ipAddress;

                textBox1.Focus();
            }
            catch (Exception ex)
            {
                EnableChatControls(false);

                HandleError(ex);
            }
        }

        private void radioButton2_CheckedChanged(object sender, EventArgs e)
        {
            Cursor.Current = Cursors.WaitCursor;

            // Attempt to run chat client
            try
            {
                EnableChatControls(true);

                RunAsClient();

                this.Text = "Chat Application - Client " + _client.ClientID +
                    " connected to " + _ipAddress;

                textBox1.Focus();
            }
            catch (Exception ex)
            {
                EnableChatControls(false);

                HandleError(ex);
            }
        }

        void EnableChatControls(bool enable)
        {
            // Disable/enable GUI controls
            radioButton1.Enabled = !enable;
            radioButton2.Enabled = !enable;
            panel1.Visible = true;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            // Send a message
            try
            {
                AppendMessage("<Me> " + textBox1.Text);

                SendMessage(textBox1.Text);

                textBox1.Text = "";
            }
            catch (Exception ex)
            {
                HandleError(ex);
            }
        }
        #endregion
    }
}